/*
 * * Copyright (c) 2013-2015, Panasonic Corporation.
 * *
 * * Permission to use, copy, modify, and/or distribute this software for any
 * * purpose with or without fee is hereby granted, provided that the above
 * * copyright notice and this permission notice appear in all copies.
 * *
 * * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * *
 * * --- TESTING DISCLAIMER ---
 * *
 * * Please note that the code supplied within this module is highly designed
 * * around working with the Panasonic Distributed Testing Framework(https://github.com/PSLCL/testing-framework).
 * * As such, there are certain coding methodologies and test design concepts throughout.
 * *
 * * Though these tests are system tests, they are designed to fail immediately
 * * if any problems arise throughout execution, for quicker retrieval of results.
 * * In a real system application, the developer should provide a means for proper handling
 * * of failures in order to maintain the system.
 * *
 * * Due to the nature of this module, the source code should NOT be used in anyway for training,
 * * learning, or as sample code for providing proper application design with OpenDOF projects.
 * * Instead if one is interested in learning about proper application design they should reference
 * * the particular SDK or training modules provided by OpenDOF.
 * * These can be found in the GIT repository of the module that this testing module is testing or
 * * you can also look at the Downloads page(https://opendof.org/downloads/) for OpenDOF.
 */

package org.opendof.core.domain.credgen.test;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.io.PrintStream;

import org.opendof.core.domain.credgen.CredGen;
import org.opendof.core.oal.DOFAuthenticator;
import org.opendof.core.oal.DOFCredentials;
import org.opendof.core.oal.DOFObjectID;
import org.opendof.core.oal.DOFUtil;

public class CredGenToolTest
{

	private static final DOFObjectID.Domain domainID = DOFObjectID.Domain.create("[6:testing.opendof.org]");
	private static final DOFObjectID.Authentication authenticationNodeID = DOFObjectID.Authentication.create("[3:aNode@testing.opendof.org]");

	private String credentialsHex = DOFUtil
			.bytesToHexString(DOFCredentials.Key.create(authenticationNodeID, DOFAuthenticator.createSessionKey()).getBytes());
	private String credentialsFileURL = "a.cred";
	private String credentialsFileURL2 = "a3.cred";
	private String credentialsFormat1 = "<1:[3:aNode@testing.opendof.org];{0000000000000000000000000000000000000000000000000000000000000000}>";
	private String credentialsFormat2 = "<2:[3:aNode@testing.opendof.org];myPassword72>";
	private String credentialsFormat1_NoSecret = "<1:[3:aNode@testing.opendof.org]>";
	private String credentialsFormat2_NoSecret = "<2:[3:aNode@testing.opendof.org]>";

	private File currentDir = new File("");
	private String workingDir;
	private static String DMCLFILEDIR = "/bin/";

	private static OutputStream systemErrOut;
	private static ByteArrayOutputStream errOut = null;
	private static OutputStream systemStdOut;
	private static ByteArrayOutputStream stdOut = null;
	private static String tempErrorStream = null;

	public static void main(String[] args)
	{
		int exitCode = 255;

		try
		{
			CredGenToolTest credGenToolTest = new CredGenToolTest();
			Runtime.getRuntime().addShutdownHook(credGenToolTest.new ShutdownHook());

			try
			{
				credGenToolTest.printCommands();

				System.out.println("<<<running>>>");
				exitCode = credGenToolTest.run();
			}
			finally
			{
				credGenToolTest.cleanUp();
			}
		}
		catch (Exception e)
		{
			System.out.println("Exception: " + e.getMessage());
			e.printStackTrace();
			exitCode = 500;
		}

		System.out.println("exitCode: " + exitCode);
		System.exit(exitCode);
	}

	private int run() throws Exception
	{

		// Get the working directory path needed for reading of the files.

		try
		{
			workingDir = currentDir.getCanonicalPath() + DMCLFILEDIR;
		}
		catch (Exception e)
		{
			// Ignore.
		}

		// Test all transactions.

		executePositiveTests();
		executeNegativeTests();

		return 0;
	}

	private void executePositiveTests() throws Exception
	{
		System.out.println("\nRunning Positive CredGen Tool Tests now...");

		System.out.println("\nTesting the generation of byte credentials by a hex string...\n");

		if (!runProcess(false, credentialsHex, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--show",
				"--force"))
			throw new Exception("Exception: [CredGen] Unable to generate byte credentials from a hex string.");

		System.out.println("Testing the generation of byte credentials by a file...\n");

		if (!runProcess(false, workingDir + credentialsFileURL, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2,
				"--show", "--force"))
			throw new Exception("Exception: [CredGen] Unable to generate byte credentials from a file.");

		System.out.println("Testing the generation of byte credentials by Credential Factory formats...\n");

		if (!runProcess(false, credentialsFormat1, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--show",
				"--force"))
			throw new Exception("Exception: [CredGen] Unable to generate byte credentials from a Key type Credential Factory format.");

		if (!runProcess(false, credentialsFormat2, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--show",
				"--force"))
			throw new Exception("Exception: [CredGen] Unable to generate byte credentials from a Password type Credential Factory format.");

		System.out.println("Testing the generation of byte credentials by Credential Factory formats that have no secret added...");

		if (!runProcess(false, credentialsFormat1_NoSecret, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2,
				"--show", "--force"))
			throw new Exception(
					"Exception: [CredGen] Unable to generate byte credentials from a Key type Credential Factory format with no secret added.");

		if (!runProcess(false, credentialsFormat2_NoSecret, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2,
				"--show", "--force"))
			throw new Exception(
					"Exception: [CredGen] Unable to generate byte credentials from a Password type Credential Factory format with no secret added.");
	}

	private void executeNegativeTests() throws Exception
	{
		System.out.println("\nRunning Negative CredGen Tool Tests now...");

		System.out.println("\nTesting no output exception...\n");

		if (runProcess(false, credentialsHex, "--domain", domainID.toStandardString(), "--file"))
			throw new Exception("Exception: [CredGen] Unable to throw no output exception.");

		System.out.println("Testing input is null exception...\n");

		if (runProcess(false, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw input is null exception.");

		System.out.println("Testing unrecognized option exception...\n");

		if (runProcess(false, credentialsHex, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force",
				"--unrecognized"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for unrecognized option using '--'.");

		if (runProcess(false, credentialsHex, "--domain", domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force", "-u"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for unrecognized option using '-'.");

		System.out.println("Testing invalid Credential format exceptions...\n");

		if (!runProcess(true, "2:[3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not beginning with '<'.");

		if (!runProcess(true, "<32768:[3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for invalid credential type (short value).");

		if (!runProcess(true, "<2[3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not delimiting the credential with ':'.");

		if (!runProcess(true, "<2:3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not beginning OID with '['.");

		if (!runProcess(true, "<2:[-1:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for OID class being negative.");

		if (!runProcess(true, "<2:[1073741824:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for OID class being greater than the max value.");

		if (!runProcess(true, "<2:[3aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not delimiting OID with ':'.");

		if (!runProcess(true, "<2:[3:aNode@testing.opendof.org;myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not ending OID with ']'.");

		if (!runProcess(true, "<2:[3:aNode@testing.opendof.org]myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not following Auth ID by a ';'.");

		if (!runProcess(true, "<2:[3:aNode@testing.opendof.org];{F6536E4EC29C7E2A90E1AC553B5B580ABBBDF55232A603B615E2FA8CFAC43FB1}>", "--domain",
				domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for a Password being a hex value.");

		if (!runProcess(true, "<1:[3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for a Key being a Password.");

		if (!runProcess(true, "<1:[3:aNode@testing.opendof.org];F6536E4EC29C7E2A90E1AC553B5B580ABBBDF55232A603B615E2FA8CFAC43FB1}>", "--domain",
				domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for a Hex string not starting with a {.");

		if (!runProcess(true, "<1:[3:aNode@testing.opendof.org];{F6536E4EC29C7E2A90E1AC553B5B580ABBBDF55232A603B615E2FA8CFAC43FB1>", "--domain",
				domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for a Hex string not ending with a }.");

		if (!runProcess(true, "<3:[3:aNode@testing.opendof.org];myPassword72>", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for unknown credential type.");

		if (!runProcess(true, "<2:[3:aNode@testing.opendof.org];myPassword72", "--domain", domainID.toStandardString(), "--file",
				workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not ending Password credentials with '>'.");

		if (!runProcess(true, "<1:[3:aNode@testing.opendof.org];{F6536E4EC29C7E2A90E1AC553B5B580ABBBDF55232A603B615E2FA8CFAC43FB1}", "--domain",
				domainID.toStandardString(), "--file", workingDir + credentialsFileURL2, "--force"))
			throw new Exception("Exception: [CredGen] Unable to throw exception for not ending Key credentials with '>'.");

		System.out.println("\nTesting no arguments exception...\n");

		if (runProcess(false))
			throw new Exception("Exception: [CredGen] Unable to throw no arguments exception.");

		System.out.println("Testing no domain ID exception...\n");

		if (runProcess(false, credentialsHex, "--file", workingDir + credentialsFileURL2, "--force", "--domain"))
			throw new Exception("Exception: [CredGen] Unable to throw no domain ID exception.");
	}

	/**
	 * Method creates a temporary output stream to catch the runtime's command line error stream. Basically, reassigns this runtime's error output
	 * stream to the created temporary output stream.
	 */
	public static void setUpCLErrOutput()
	{
		systemErrOut = System.err;
		errOut = new ByteArrayOutputStream();
		System.setErr(new PrintStream(errOut));
	}

	/**
	 * Method creates a temporary output stream to catch the runtime's command line standard output stream. Basically, reassigns this runtime's
	 * standard output stream to the created temporary output stream.
	 */
	public static void setUpCLStdOutput()
	{
		systemStdOut = System.out;
		stdOut = new ByteArrayOutputStream();
		System.setOut(new PrintStream(stdOut));
	}

	/**
	 * Method clears out the temporary output stream that held this runtime's command line error stream. It also, reverts this runtime's command line
	 * error stream to it's system output error stream.
	 */
	public static void tearDownCLErrOutput()
	{
		errOut.reset();
		errOut = null;
		System.setErr(new PrintStream(systemErrOut));
	}

	/**
	 * Method clears out the temporary output stream that held this runtime's command line standard output stream. It also, reverts this runtime's
	 * command line standard output stream to it's system output standard stream.
	 * 
	 * @param reset
	 *            Whether or not to reset the temporary output stream.
	 */
	public static void tearDownCLStdOutput(boolean reset)
	{
		if (reset)
		{
			stdOut.reset();
			stdOut = null;
		}
		System.setOut(new PrintStream(systemStdOut));
	}

	public static boolean verifyException(String error)
	{
		if (error.contains("Credential format must begin with <"))
			return true;
		else if (error.contains("Invalid credential type"))
			return true;
		else if (error.contains("Credential must be delimited by a :"))
			return true;
		else if (error.contains("Auth ID must be followed by a ;"))
			return true;
		else if (error.contains("Password can not be a hex value"))
			return true;
		else if (error.contains("Hex string must start with a {"))
			return true;
		else if (error.contains("Unknown credential type"))
			return true;
		else if (error.contains("Credential format must end with a >"))
			return true;
		else if (error.contains("Secret can not be specified"))
			return true;
		else if (error.contains("String is not hex or a format OID"))
			return true;
		else if (error.contains("Format OID must start with a ["))
			return true;
		else if (error.contains("ID Class must be positive and less than DOFObjectID.MAX_CLASS"))
			return true;
		else if (error.contains("Format OID must be delimited with a :"))
			return true;
		else if (error.contains("Format OID must end with a ]"))
			return true;
		else if (error.contains("Invalid number format"))
			return true;
		else if (error.contains("Unexpected end of string"))
			return true;
		else if (error.contains("Secret can not be specified") || error.contains("Illegal hex char: >"))
			return true;
		else
			return false;
	}

	/**
	 * Method runs a CredGen tool transaction using the command line client, then verify's whether the process was successful.
	 * 
	 * @param boolean
	 *            isTestExcept Whether or not this run process will be testing a particular exception to be thrown.
	 * @param String
	 *            [] args Represents the command line arguments used to run the process.
	 * @return true If the process completed successfully given the type of transaction and desired outcome, false otherwise.
	 */
	private static boolean runProcess(boolean isTestExcept, String... args)
	{

		// Run the process and check the command line output for errors,
		// if an exception is thrown by the command line client (during Positive Tests),
		// then the transaction failed.

		setUpCLErrOutput();
		setUpCLStdOutput();

		if (args.length == 0)
		{
			args = null;
		}

		try
		{
			CredGen.execute(args);
		}
		catch (Exception e)
		{
			// Ignore.
		}

		if (!isTestExcept)
		{
			if (errOut.toString().equals("") || errOut.toString().equals(null))
			{
				tearDownCLErrOutput();
				if (stdOut.toString().equals("") || stdOut.toString().equals(null))
				{
					tearDownCLStdOutput(true);
					return false;
				}
				else
				{
					tearDownCLStdOutput(false);
					System.out.println(stdOut.toString());
					tearDownCLStdOutput(true);
					return true;
				}
			}
			else
			{
				tearDownCLStdOutput(false);
				System.out.println(errOut.toString());
				tearDownCLStdOutput(true);
				tearDownCLErrOutput();
				return false;
			}
		}
		else
		{
			if (!errOut.toString().equals("") || !errOut.toString().equals(null))
			{
				tempErrorStream = errOut.toString();
				tearDownCLErrOutput();
				tearDownCLStdOutput(true);
				if (verifyException(tempErrorStream))
					return true;
				else
					return false;
			}
			else
			{
				tearDownCLErrOutput();
				return false;
			}
		}
	}

	private void cleanUp()
	{
		System.out.println("clean up...");

		if (errOut != null)
		{
			tearDownCLErrOutput();
		}
		if (stdOut != null)
		{
			tearDownCLStdOutput(true);
		}
		System.out.println("<<<stopped>>>");
	}

	public void printCommands()
	{
		String commands = "Commands and Defaults:\n";
		commands += "\tcredentialsFiles:	 " + credentialsFileURL + ", " + credentialsFileURL2 + "\n";
		System.out.println(commands);
	}

	public class ShutdownHook extends Thread
	{
		@Override
		public void run()
		{
			cleanUp();
		}
	}

}
